import torch
import cupy as cp

DEVICE = cp.cuda.Device().id

@cp.memoize(for_each_device=True)
def cunnex(func_name, func_body):
  return cp.cuda.compile_with_cache(func_body).get_function(func_name)

class Stream:
  def __init__(self, ptr):
    self.ptr = ptr
  
class CudaCallable:
  def __init__(self):
    self._use_torch_in_cupy_malloc()
    self.device = torch.device(DEVICE)
    self.stream = Stream(torch.cuda.current_stream(device=DEVICE).cuda_stream)
    self.kernel = ""

  @staticmethod
  def _torch_alloc(size):
    tensor = torch.empty(size, dtype=torch.uint8, device=DEVICE)
    return cp.cuda.MemoryPointer(
        cp.cuda.UnownedMemory(tensor.data_ptr(), size, tensor), 0)

  def _use_torch_in_cupy_malloc(self):
    cp.cuda.set_allocator(self._torch_alloc)

  def _compile_kernel_str(
      self,
      kernel,
      name,
      options=(),
      backend="nvrtc",
      max_dynamic_smem=None
    ):
    fn = cp.RawKernel(
      kernel,
      name,
      options=options,
      backend=backend,
    )
    if max_dynamic_smem:
      fn.max_dynamic_shared_size_bytes = max_dynamic_smem
    
    # method_name = f"_{name}"
    # assert not hasattr(self, method_name), f"kernel {method_name} already exists"
    # setattr(self, method_name, fn)
    return fn


import numpy as np
import torch
from datetime import datetime
from pathlib import Path
import math
import psutil
from os import path

def sliding_window(array, window_size, start_at=0, max_count=None):
  if max_count is None:
    max_count = len(array)
  start = start_at + 1 - window_size + 1
  
  sub_windows = (
    start +
    # expand_dims are used to convert a 1D array to 2D array.
    np.expand_dims(np.arange(window_size), 0) +
    np.expand_dims(np.arange(max_count + 1), 0).T
  ) % len(array)
  return torch.tensor(array[sub_windows], dtype=torch.long)

class TableLogger:
  def __init__(self, log_dir, col_labels, col_types=None, col_lengths=None):
    self.log_dir = Path(log_dir)
    now = datetime.now()
    self.file_path = self.log_dir / f"{now.date()}_{now.hour}_{now.minute}.log"
    self.n_cols = len(col_labels)
    if col_types is None:
      col_types = [str for i in range(self.n_cols)]
    
    if col_lengths is None:
      # col_lengths = [15 for i in range(self.n_cols)]
      col_lengths = [max(3, len(label)) for label in col_labels]
    assert len(col_labels) == len(col_types) == len(col_lengths)

    self.col_labels = col_labels
    self.col_types = col_types
    self.col_lengths = col_lengths
    self.format_str = [f"{{:<{col_lengths[i]}.{col_lengths[i]}}}" for i in range(self.n_cols)]
    self.format_str = f"| {' | '.join(self.format_str)} |\n"
    # self.title_line = self.format_str.format(*self.col_labels)
    # self.dash_line = self.format_str.format(*["-"*i for i in self.col_lengths])

    self.add_row(self.col_labels, remember_types=False)
    self.add_row(["-"*i for i in self.col_lengths], remember_types=False)

  def add_row(self, row_items, remember_types=True):
    assert len(row_items) == self.n_cols
    if remember_types:
      self.col_types = [type(item) for item in row_items]
    line = self.format_str.format(*[str(item) for item in row_items])
    with open(self.file_path, "a") as f:
      f.write(line)

  def read_row(self, row):
    items = row.split("|")[1:-1]
    if len(items) != self.n_cols:
      return None
    
    # items = [item.strip() for item in items]
    items = [items[i].strip() for i in range(self.n_cols)]
    items = [self.col_types[i](items[i]) for i in range(self.n_cols)]
    return items

  def read_table(self):
    with open(self.file_path, "r") as f:
      rows = f.read().split("\n")[2:]
    row_items = [self.read_row(row) for row in rows]
    row_items = [i for i in row_items if i is not None]
    return row_items

def str2dtype(s):
  if isinstance(s, torch.dtype):
    return s
  elif isinstance(s, str):
    if s in ["bool"]:
      return torch.bool
    elif s in ["uint8", "char"]:
      return torch.uint8
    elif s in ["int8", "byte"]:
      return torch.int8
    elif s in ["int16", "short"]:
      return torch.int16
    elif s in ["int32", "int"]:
      return torch.int32
    elif s in ["int 64", "long"]:
      return torch.int64
    elif s in ["float16", "half"]:
      return torch.float16
    elif s in ["bfloat16"]:
      return torch.bfloat16
    elif s in ["float32", "float", ]:
      return torch.float32
    elif s in ["float64", "double"]:
      return torch.float64
    else:
      raise ValueError(f"unrecognized type string {s}")

  else:
    raise RuntimeError

def dtype2ctype(dtype):
  dtype = str2dtype(dtype)
  if dtype == torch.bool:
    return "bool"
  elif dtype == torch.uint8:
    return "uint8_t"
  elif dtype == torch.int8:
    return "int8_t"
  elif dtype == torch.int16:
    return "int16_t"
  elif dtype == torch.int32:
    return "int32_t"
  elif dtype == torch.int64:
    return "int64_t"
  elif dtype == torch.float16:
    return "half"
  elif dtype == torch.bfloat16:
    return "bfloat16"
  elif dtype == torch.float32:
    return "float"
  elif dtype == torch.float64:
    return "double"

def next_power_of_2(n):
    count = 0
    # First n in the below
    # condition is for the
    # case where n is 0
    if (n and not(n & (n - 1))):
        return n
     
    while( n != 0):
        n >>= 1
        count += 1
     
    return 1 << count

def unique_first_occurrence(x=None, unique=None, inverse=None, return_indices=False, dim=None):
  if unique is None and inverse is None:
    unique, inverse = torch.unique(x, dim=dim, sorted=True, return_inverse=True)
  elif x is None:
    unique, inverse = unique, inverse
  else:
    raise RuntimeError("must provide x or unique and inverse")

  perm = torch.arange(inverse.size(0), dtype=inverse.dtype, device=inverse.device)
  inverse_flipped, perm = inverse.flip([0]), perm.flip([0])
  perm = inverse_flipped.new_empty(unique.size(0)).scatter_(0, inverse_flipped, perm)
  if return_indices:
    return perm
  else:
    mask = torch.zeros_like(inverse, dtype=torch.bool)
    mask[perm] = True    
    return mask 

def expand_tensor(tensor, dim=0, fill=0, by=1.0):
  shape = list(tensor.shape)
  n_dims = len(shape)
  assert n_dims > 0, "tensor needs to have at least 1 dimension"
  if isinstance(by, float):
    shape[dim] = math.ceil(shape[dim] * by)
  elif isinstance(by, int):
    shape[dim] = by
  
  new_tensor = torch.zeros(shape, dtype=tensor.dtype, device=tensor.device)
  new_tensor.fill_(fill)
  new_tensor = torch.cat([tensor, new_tensor], dim=dim)
  return new_tensor

def batch_allclose(a, b, tol=1e-6):
  a, b = torch.broadcast_tensors(a, b)
  assert a.dtype == b.dtype
  assert a.device == b.device
  assert a.shape == b.shape
  assert len(a.shape) > 1
  batch_size = a.shape[0]
  a = a.view(batch_size, -1)
  b = b.view(batch_size, -1)
  if a.dtype in {torch.int, torch.long, torch.int16, torch.int8, torch.uint8, torch.bool}:
    return (a != b).sum(dim=-1) == 0

  elif a.dtype in {torch.half, torch.float, torch.double, torch.bfloat16}:
    return ((a - b).abs() >= tol).sum(dim=-1) == 0

def check_available_ram(device="cpu"):
  """
  Returns available RAM on target device
  args:
    device:     str or torch.device
  """
  if isinstance(device, str):
    device = torch.device(device)
  elif isinstance(device, torch.device):
    device = device
  else:
    raise RuntimeError("`device` must be str or torch.device")
  
  if device.type == "cpu":
    return psutil.virtual_memory().available
  else:
    total = torch.cuda.get_device_properties(device).total_memory
    used = torch.cuda.memory_allocated(device)
    return total - used

def will_it_fit(size, device="cpu", safe_mode=True):
  """
  Returns True if an array of given byte size fits in target device.
  if self.safe_mode = False, this function simply compares the given byte size with the remaining RAM on target device. This option is faster, 
      but it doesn't take memory fragmentation into account. So it will still be possible to run out of memory.
  if self.safe_mode = True, it will try to allocate a tensor with the given size. if allocation fails, return False. 
      This option is recommended when the other option fails because of OOM.
  
  args:
      size:       int
      device:     str or torch.device
      safe_mode:  bool
  returns:
      result:     bool
  """
  if safe_mode:
    try:
      torch.empty(size, device=device, dtype=torch.uint8)
    except:
      return False
    return True
  else:
    return check_available_ram(device) >= size

def find_optimal_splits(n, get_required_memory, device="cpu", safe_mode=True):
  """
  Find an optimal number of split for `n`, such that `get_required_memory(math.ceil(n / n_split))` fits in target device's RAM.
  get_required_memory should be a fucntion that receives `math.ceil(n/n_split)` and returns the required memory in bytes.
  args:
    n:                      int
    get_required_memory:    function
    device:                 str or torch.device
    safe_mode:              bool
  returns:
    n_splits:               int
  """
  splits = 1
  sub_n = n
  break_next = False
  while True:
    if break_next:
      break
    if splits > n:
      splits = n
      break_next = True
    sub_n = math.ceil(n / splits)
    required_memory = get_required_memory(sub_n)
    if will_it_fit(required_memory, device):
      break
    else:
      splits *= 2
      continue
  return splits

def get_absolute_path(*relative_path):
  relative_path = path.join(*relative_path)
  return path.join(path.dirname(__file__), relative_path)

def merge(candidate_pairs, selected_pair_idx, replacement):
  """
    candidate_pairs : [2, n_seqs, new_seq_len]
    selected_pair_idx : [n_seqs]
    replacement : [n_seqs]
  """
  n_seqs = selected_pair_idx.shape[0]
  new_seq_len = candidate_pairs.shape[2]
  assert candidate_pairs.shape == (2, n_seqs, new_seq_len)
  assert selected_pair_idx.shape == (n_seqs,)
  assert replacement.shape == (n_seqs,)
  arange = torch.arange(new_seq_len)[None]
  # print(arange.shape)
  # print(selected_pair_idx[:, None].shape)
  # print(candidate_pairs.shape)
  # print(candidate_pairs.shape)
  # print(selected_pair_idx.shape)
  # print((arange < selected_pair_idx).shape )

  merged_seq = candidate_pairs[0] * (arange < selected_pair_idx[:, None])
  merged_seq += candidate_pairs[1] * (arange > selected_pair_idx[:, None])
  merged_seq += replacement[:, None] * (arange == selected_pair_idx[:, None])
  return merged_seq



import torch
import torch.nn.functional as F
from typing import Union, Optional
from torch import Tensor

from .components.cuda.topkspspcsim import TopkSPSPCSIM


def support_dense(func):
  func_name = func.__name__
  if hasattr(torch, func_name):
    torch_func = getattr(torch, func_name)

  elif hasattr(F, func_name):
    torch_func = getattr(F, func_name)

  else:
    return func

  def new_func(*args, **kwargs):
    all_tensors_in_args = [i for i in args if isinstance(i, Tensor)]
    all_tensors_in_kwargs = [i for i in kwargs.values() if isinstance(i, Tensor)]
    all_tensors = all_tensors_in_args + all_tensors_in_kwargs
    all_dense = all([not tensor.layout in {torch.sparse_coo, torch.sparse_csr} for tensor in all_tensors])
    if all_dense:
      return torch_func(*args, **kwargs)
    return func(*args, **kwargs)

  return new_func

@support_dense
def sum(
    input,
    dim = None, 
    keepdim = False,
  ):
  # if not tensor.is_sparse:
  #   return torch.sum(tensor, dim=dim, keepdim=keepdim)

  y = torch.sparse.sum(input, dim=dim)
  
  if keepdim:
    if isinstance(dim, int):
      y = y.unsqueeze(dim=dim)

    elif isinstance(dim, (tuple, list)):
      dim = sorted(dim, reverse=False)
      for d in dim:
        y = y.unsqueeze(dim=d)

    elif isinstance(dim, None):
      for d in range(input.ndim):
        y.unsqueeze(dim=0)

  return y

@support_dense
def norm(
    input, 
    dim = None,
    p = 2, 
    keepdim = False,
  ):
  # if not tensor.is_sparse:
  #   return torch.norm(tensor, dim=dim, p=p, keepdim=keepdim)

  if p == 0:
    new_values = input.values().bool().to(input.dtype)
  elif p == 1:
    new_values = input.values().abs()
  elif p == 2:
    new_values = input.values().square()
  else:
    new_values = input.values().abs().pow(p)
  if input.layout == torch.sparse_coo:
    input = torch.sparse_coo_tensor(
      indices=input.indices(),
      values=new_values,
      size=input.shape,
      dtype=input.dtype,
      device=input.device,
    )
  elif input.layout == torch.sparse_csr:
    torch.sparse_csr_tensor(
      crow_indices=input.crow_indices(),
      col_indices=input.col_indices(),
      values=new_values,
      size=input.shape,
      dtype=input.dtype,
      device=input.device
    )
  y = input.sum(dim=dim, keepdim=keepdim)
  if p == 2:
    y = y.sqrt()
  else:
    y = y.pow(1 / p)

  return y

@support_dense
def softmax(
    input, 
    dim = -1,
  ):
  dim = dim % input.ndim
  # if not tensor.is_sparse:
  #   return torch.softmax(tensor, dim=dim)

  return torch.sparse.softmax(input, dim=dim)

@support_dense
def log_softmax(
    input, 
    dim = -1,
  ):
  dim = dim % input.ndim
  # if not tensor.is_sparse:
  #   return torch.log_softmax(tensor, dim=dim)

  return torch.sparse.log_softmax(input, dim=dim)

@support_dense
def normalize(
    input,
    dim = 0,
    p = 2,
    eps = 1e-12,
  ):
  # if not input.is_sparse:
  #   return F.normalize(input, dim=dim, p=p, eps=eps)
  dim = dim % input.ndim
  if input.layout == torch.sparse_csr:
    input = input.to_sparse_coo()
    
  if not input.is_coalesced():
    input = input.coalesce()
  
  if p == 0:
    new_values = input.values().bool().to(input.dtype)
  elif p == 1:
    new_values = input.values().abs()
  elif p == 2:
    new_values = input.values().square()
  else:
    new_values = input.values().abs().pow(p)
  new_values.clamp_min_(eps)
  new_values = new_values.log()

  input = torch.sparse_coo_tensor(
    indices=input.indices(),
    values=new_values,
    size=input.shape,
    dtype=input.dtype,
    device=input.device,
  )

  y = softmax(input, dim=dim)
  if p == 2:
    return y.sqrt()
  else:
    return y.pow(1 / p)

def normalize_(
    input,
    dim = -1,
    p  = 2,
    eps = 1e-12,
  ):
  dim = dim % input.ndim
  if input.is_sparse or input.is_sparse_csr:
    y = normalize(input=input, dim=dim, p=p, eps=eps)
    input.values()[:] = y.values()
  else:
    input_norm = input.norm(p=p, dim=dim, keepdim=True) + eps
    input.div_(input_norm)
  return input

def freq2prob(freqs, dim=-1):
  return normalize(freqs, dim=dim, p=1)

def cross_entropy(probs, labels, reduction="mean"):
  """
    probs : Tensor, shape = [n, v]
    labels : Tensor, shape = [n]
  """
  ce = -probs.gather(index=labels[:, None], dim=-1).log().flatten()
  # print(f"{ce=}")
  if reduction == "sum":
    ce = ce.sum()
  elif reduction == "mean":
    ce = ce.mean()
  elif reduction == "none":
    pass

  return ce

topkspsp_inner = None
topkspsp_nl1 = None
topkspsp_nl2 = None

def sparse_topk_inner(a, b, dim=1, k=1):
  assert b.ndim == a.ndim == 2
  dim = dim % a.ndim
  assert dim == 1
  assert a.layout in {torch.sparse_coo, torch.sparse_csr}
  assert b.layout in {torch.sparse_coo, torch.sparse_csr}
  global topkspsp_inner
  if topkspsp_inner is None:
    topkspsp_inner = TopkSPSPCSIM(sim_type="inner")
  return topkspsp_inner(a, b, k=k)

def sparse_topk_nl1(a, b, dim=1, k=1):
  assert b.ndim == a.ndim == 2
  dim = dim % a.ndim
  assert dim == 1
  assert a.layout in {torch.sparse_coo, torch.sparse_csr}
  assert b.layout in {torch.sparse_coo, torch.sparse_csr}
  global topkspsp_nl1
  if topkspsp_nl1 is None:
    topkspsp_nl1 = TopkSPSPCSIM(sim_type="nl1")
  return topkspsp_nl1(a, b, k=k)

def sparse_topk_nl2(a, b, dim=1, k=1):
  assert b.ndim == a.ndim == 2
  dim = dim % a.ndim
  assert dim == 1
  assert a.layout in {torch.sparse_coo, torch.sparse_csr}
  assert b.layout in {torch.sparse_coo, torch.sparse_csr}
  global topkspsp_nl2
  if topkspsp_nl2 is None:
    topkspsp_nl2 = TopkSPSPCSIM(sim_type="nl2")
  return topkspsp_nl2(a, b, k=k)

def sparse_topk_cos(a, b, dim=1, k=1):
  assert b.ndim == a.ndim == 2
  dim = dim % a.ndim
  assert dim == 1
  assert a.layout in {torch.sparse_coo, torch.sparse_csr}
  assert b.layout in {torch.sparse_coo, torch.sparse_csr}
  a = normalize(a, dim=-1)
  b = normalize(b, dim=-1)
  global topkspsp_inner
  if topkspsp_inner is None:
    topkspsp_inner = TopkSPSPCSIM(sim_type="inner")
  return topkspsp_inner(a, b, k=k)



import torch
import numpy as np
from torch import Tensor

from .cuda_closed_hashmap import CudaClosedHashmap
from ..util import str2dtype, next_power_of_2
from .. import functions as fs

class SparseDOKTensor(object):
  def __init__(self, size, indices=None, values=None, device=None, dtype=None):
    assert isinstance(size, (list, tuple))
    self._size = torch.Size(size)
    
    if device is None:
      if indices is not None:
        self.device = indices.device
      else:
        self.device = "cuda:0"
    else:
      self.device = torch.device(device)

    if dtype is None:
      if values is not None:
        self._dtype = values.dtype
      else:
        self._dtype = torch.float32
    else:
      self._dtype = str2dtype(dtype)

    if indices is None:
      self._hashmap = CudaClosedHashmap(
        n_buckets=32,
        device=self.device,
        # subkey_inds=self.subkey_inds,
      )
    else:
      assert values is not None
      assert len(indices.shape) == 2
      assert len(values.shape) == 1
      assert indices.shape[1] == values.shape[0]
      self._hashmap = CudaClosedHashmap(
        n_buckets=next_power_of_2(values.shape[0]),
        device=self.device,
        # subkey_inds=self.subkey_inds,
      )
      self._hashmap.set(
        indices.T.to(self.device),
        values.to(device=self.device, dtype=self.dtype)[:, None]
      )

  @classmethod
  def from_sparse_coo(cls, tensor):
    assert isinstance(tensor, Tensor)
    assert tensor.layout == torch.sparse_coo
    tensor = tensor.coalesce()
    indices = tensor._indices()
    values = tensor._values()

    return cls(
      size=tensor.shape,
      indices=indices, 
      values=values, 
      device=tensor.device
    )

  @classmethod
  def from_dense(cls, tensor):
    assert isinstance(tensor, Tensor)
    assert tensor.layout == torch.strided
    return cls.from_sparse_coo(tensor.to_sparse())

  @classmethod
  def __torch_function__(cls, func, types, args=(), kwargs=None):
    if kwargs is None:
      kwargs = {}

    args = [arg.to_sparse_coo() if isinstance(arg, cls) else arg for arg in args]
    kwargs = {key: (arg.to_sparse_coo() if isinstance(arg, cls) else arg) for key, arg in kwargs.items()}

    return func(*args, **kwargs)
    
  @property
  def is_sparse(self):
    return True

  def _is_coalesced(self):
    return False

  def size(self):
    return self._size

  @property
  def shape(self):
    return self.size()

  @property
  def dtype(self):
    return self._dtype

  def clone(self, deepclone=False):
    new =  SparseDOKTensor(
      size=self._size, 
      indices=None, 
      values=None, 
      device=self.device, 
      dtype=self.dtype, 
    )
    new._hashmap = self._hashmap.clone(deepclone=deepclone)
    return new

  def indices(self):
    # assert self._hashmap.keys() is not None
    indices = self._hashmap.keys()
    if indices is None:
      return torch.empty(self.ndim, 0, device=self.device, dtype=torch.long)
    else:
      return indices.T

  def values(self):
    values = self._hashmap.values()
    if values is None:
      return torch.empty(0, device=self.device, dtype=self.dtype)
    else:
      return values.flatten()

  def set_values(self, selectors, new_values):
    self._hashmap.set_values(selectors=selectors, new_values=new_values[:, None])

  def _indices(self):
    return self._hashmap._keys

  def _values(self):
    return self._hashmap._values
  
  @property
  def ndim(self):
    return len(self.shape)

  def _nnz(self):
    return self._hashmap.n_elements

  @property
  def sparsity(self):
    return 1 - self._nnz() / float(np.prod(self.shape))

  @property
  def _as_sparse_coo_tensor(self):
    return self.to_sparse_coo()

  def to_sparse_coo(self): 
    t = torch.sparse_coo_tensor(
      indices=self.indices(), 
      values=self.values(),
      size=self.shape,
      device=self.device,
      )
    # t._coalesced_(True)
    t = t.coalesce()
    return t

  def to_sparse_csr(self):
    return self.to_sparse_coo().to_sparse_csr()

  def to_dense(self):
    return self.to_sparse_coo().to_dense()
  
  def resize(self, *new_size):
    new_size = new_size + (self.ndim - len(new_size)) * (-1, )
    assert len(new_size) == self.ndim
    new_size = torch.Size([self.shape[i] if new_size[i] == -1 else new_size[i] for i in range(self.ndim)])
    self._size = new_size

  def is_coalesced(self):
    return self._is_coalesced

  def _handle_selectors(self, selectors):
    if isinstance(selectors, (Tensor, slice) ):
      selectors = (selectors, )

    _selectors = []
    for selector in selectors:
      if isinstance(selector, Tensor) and selector.dtype == torch.bool:
        _selectors += selector.nonzero(as_tuple=True)

      elif isinstance(selector, Tensor) and selector.dtype in {torch.int64, torch.int32, torch.int16}:
        _selectors.append(selector.long())
      
      elif isinstance(selector, np.ndarray) and selector.dtype == np.bool8:
        selector = torch.from_numpy(selector, device=self.device)
        _selectors.append(selector.nonzero(as_tuple=True))

      elif isinstance(selector, np.ndarray) and selector.dtype in {np.int16, np.int32, np.int64}:
        selector = torch.from_numpy(selector, device=self.device, dtype=torch.long)
        _selectors.append(selector)

      elif isinstance(selector, int):
        _selectors.append(torch.tensor(selector, device=self.device, dtype=torch.long))

      elif isinstance(selector, slice):
        _selectors.append(selector)

      elif isinstance(selector, (tuple, list)):
        selector = torch.tensor(selector, device=self.device, dtype=torch.long)
        _selectors.append(selector)

      else:
        raise ValueError
    selectors = _selectors + [slice(None)] * (self.ndim - len(_selectors))
    slices = [selector for selector in selectors if isinstance(selector, slice)]
    if len(selectors) != len(slices):
      broadcasted_shape = torch.broadcast_shapes(*[index.shape for index in selectors if isinstance(index, Tensor)])
    else:
      broadcasted_shape = tuple()

    selector_ndim = len(broadcasted_shape) + len(slices)
    # broadcasted_shape = torch.broadcast_shapes(*[index.shape for index in selectors if isinstance(index, Tensor)])
    try:
      first_tensor_idx = [type(i) for i in selectors].index(Tensor)
    except ValueError:
      first_tensor_idx = None

    idx = 0
    first_tensor_met = False
    for i in range(len(selectors)):
      if isinstance(selectors[i], Tensor):
        newaxis = [None] * selector_ndim
        for j in range(first_tensor_idx, first_tensor_idx + len(broadcasted_shape)):
          newaxis[j] = slice(None)
        selectors[i] = selectors[i][newaxis]
        if not first_tensor_met:
          idx += len(broadcasted_shape)
          first_tensor_met = True

      elif isinstance(selectors[i], slice):
        start, stop = selectors[i].start, selectors[i].stop
        start = 0 if start is None else start
        stop = self.shape[i] if stop is None else stop
        stop = max(min(stop, self.shape[i]), start)
        # stop = max(start, self.shape[i] if stop is None else stop)
        newaxis = [None] * selector_ndim
        newaxis[idx] = slice(None)
        selectors[i] = torch.arange(start, stop, device=self.device)[newaxis]
        idx += 1
    selectors = torch.stack(torch.broadcast_tensors(*selectors))
    return selectors

  def __getitem__(self, selectors):
    selectors = self._handle_selectors(selectors)
    values_shape = selectors.shape[1:]
    selectors = selectors.view(self.ndim, -1)

    if self._hashmap._values is None:
      values = torch.zeros(selectors.shape[1:], device=selectors.device, dtype=self.dtype)
    else:
      values, _ = self._hashmap.get(selectors.T, fallback_value=0)
    values = values.view(values_shape)
    return values

  def __setitem__(self, selectors, values):
    if not isinstance(values, Tensor):
      values = torch.tensor(values, device=self.device, dtype=self.dtype)
    else:
      assert values.dtype == self.dtype

    selectors = self._handle_selectors(selectors).unbind(dim=0)
    broadcasted = torch.broadcast_tensors(*selectors, values)
    selectors = torch.stack(broadcasted[:-1]).view(self.ndim, -1).T.contiguous()
    values = broadcasted[-1].view(-1)[:, None].contiguous()
    
    mask = values.squeeze(-1) != 0
    values = values[mask, :]
    selectors = selectors[mask, :]
    self._hashmap.set(selectors, values)

  def _get_slice_mask(self, slice, dim=0):
    start, stop, step = slice.start, slice.stop, slice.step
    if step is not None or step != 1:
      raise RuntimeError("only step size of 1 is supported")

    if start is None:
      start = 0

    if stop is None:
      stop = self.nnz

    if stop < start:
      stop = start

    indices = self.indices[dim]
    return (start <= indices) & (indices < stop)

  def _get_slice(self, slices):
    slices = tuple(slices) + (None,) * (self.ndim - len(slices))
    mask = torch.ones(self.nnz, device=self.device, dtype=torch.bool)
    new_size = []
    for dim, slice in enumerate(slices):
      if slice is None:
        new_size.append(self.shape[dim])
        continue
      new_size.append(max(slice.stop - slice.start, 0))
      mask = mask & self._get_slice_mask(slice, dim=dim)

    final_indices = self.indices[:, mask]
    final_values = self.values[mask]
    new_view = SparseDOKTensor(indices=final_indices, values=final_values, size=new_size, device=self.device, dtype=self.dtype)
    return new_view

  def norm(self, dim=None, p=2, keepdim=False):
    return fs.norm(
      input=self.to_sparse_coo(), 
      dim=dim, 
      p=p, 
      keepdim=keepdim
    )

  def sum(self, dim=None, keepdim=False):
    return fs.sum(
      input=self.to_sparse_coo(), 
      dim=dim, 
      keepdim=keepdim
    )

  def storage(self):
    return self._hashmap

  def normalize(self, dim = -1, p = 2, eps = 1e-12):
    return fs.normalize(
      input=self.to_sparse_coo(),
      dim=dim,
      p=p,
      eps=eps
    )
    
  def normalize_(self, dim = -1, p = 2, eps = 1e-12):
    return fs.normalize_(
      input=self.to_sparse_coo(),
      dim=dim,
      p=p,
      eps=eps
    )

  def softmax(self, dim=None):
    return fs.softmax(
      input=self.to_sparse_coo(), 
      dim=dim, 
    )

  def log_softmax(self, dim=None):
    return fs.log_softmax(
      input=self.to_sparse_coo(), 
      dim=dim, 
    )

  def index_select(self, dim, index):
    return self.to_sparse_coo().index_select(dim=dim, index=index)
  
  def select(self, dim, index):
    return self.to_sparse_coo().select(dim=dim, index=index)

  def permute_(self, *ordering):
    # self._hashmap.permute_keys(ordering)
    self._hashmap.key_perm = ordering
    self._size = torch.Size(self._size[i] for i in ordering)
    return self
  
  def permute(self, *ordering):
    new = self.clone()
    # new._hashmap.permute_keys(ordering)
    new._hashmap.key_perm = ordering
    new._size = torch.Size(new._size[i] for i in ordering)
    return new

  def transpose_(self, dim0, dim1):
    assert dim0 != dim1
    dim0 = dim0 % self.ndim
    dim1 = dim1 % self.ndim
    ordering = np.arange(self.ndim)
    ordering[dim0] = dim1
    ordering[dim1] = dim0
    return self.permute_(*ordering)

  def transpose(self, dim0, dim1):
    # return self.to_sparse_coo().transpose(dim0=dim0, dim1=dim1)
    assert dim0 != dim1
    dim0 = dim0 % self.ndim
    dim1 = dim1 % self.ndim
    ordering = np.arange(self.ndim)
    ordering[dim0] = dim1
    ordering[dim1] = dim0
    return self.permute(*ordering)

  def t(self):
    assert self.ndim == 2, "tensor isn't 2 dimensional, try x.transpose(dim1, dim2)"
    return self.transpose(0, 1)
  
  @property
  def T(self):
    return self.t()





from collections import namedtuple
import cupy as cp
import torch
import numpy as np
import math
from torch import Tensor
from typing import Union
import sympy
from torchtimer import ProfilingTimer

from ..cuda_callable import CudaCallable
from ...util import get_absolute_path

ProblemSize = namedtuple("ProblemSize", ["m", "n", "k"])

class TopkSPSPCSIMCuda(CudaCallable):
  """
    Topk sparse sparse cross similarity kernel
  """
  def __init__(self, tpb=256, maxnnzpr=8, minnnzpr=0, tile_m=1, stack_cap=2, bin_search_ver=3, threads_per_group=32, sim_type="inner"):
    super().__init__()
    assert 32 <= tpb <= 1024
    self.tpb = tpb
    self.maxnnzpr = maxnnzpr
    self.minnnzpr = minnnzpr
    self.tile_m = tile_m
    self.stack_cap = stack_cap
    self.num_buckets = self.tile_m * self.maxnnzpr
    self.bin_search_ver = bin_search_ver
    self.threads_per_group = threads_per_group

    self.p = 1.0
    if sim_type in {"dot", "inner"}:
      self.sim_type = "SIM_INNER"

    elif sim_type in {"nl2", "negative_euclidean"}:
      self.sim_type = "SIM_NL2"
    
    elif sim_type in {"nl1", "negative_manhattan"}:
      self.sim_type = "SIM_NL1"
    
    elif sim_type.startswith("nl"):
      self.sim_type = "SIM_NLP"
      self.p = float(sim_type.replace("nl", ""))
    
    else:
      raise NotImplementedError(f"unsupported sim_fn {sim_type}")

    self.smem_size = self.tpb * 12 + self.tile_m * 8 + self.num_buckets * 12
    assert self.smem_size <= cp.cuda.runtime.getDeviceProperties(self.device.index)["sharedMemPerBlock"]

    cu_files = [
      "head.cu",
      "bitonic.cu",
      "smem_tensor.cu",
      "stack.cu",
      "binary_search.cu",
      # "smem_hashmap.cu",
      "reduce.cu",
      "topkspspcsim.cu",
    ]
      
    kernel = []
    for file in cu_files:
      with open(get_absolute_path("components", "cuda", file), "r") as f:
        kernel.append(f.read())

    self.kernel = "\n".join(kernel)

    with open(get_absolute_path("components", "cuda", "preview.cu"), "w") as f:
      f.write(self.kernel)

    self.kernel = (self.kernel
      .replace("_TPB_", str(self.tpb))
      .replace("_MAXNNZPR_", str(self.maxnnzpr))
      .replace("_TILEM_", str(self.tile_m))
      .replace("_STACKCAP_", str(self.stack_cap))
      .replace("_BINSEARCHVER_", str(self.bin_search_ver))
      .replace("_TPG_", str(self.threads_per_group))
      .replace("_SIMTYPE_", str(self.sim_type))
      .replace("_P_", str(self.p))
      .replace("int64_t", "ll_t")
    )

    self.fn = cp.RawKernel(
      self.kernel,
      "topkspspcsim",
      backend="nvcc",
      options=(
        '-std=c++17',
        "--device-c",
      )
    )
    self.fn.max_dynamic_shared_size_bytes = self.smem_size
    # print(self.fn.attributes)
    
  def __call__(
      self, 
      a : Tensor,  #[m, k]
      b : Tensor, #[n]
      n_candidates : int = 256,
    ):
    assert isinstance(a, Tensor)
    assert a.layout in {torch.sparse_csr}
    assert 1 <= n_candidates <= self.tpb
    m, k = a.shape
    n, _ = b.shape
      
    a_crow_inds = a.crow_indices()
    a_row_start = a_crow_inds[:-1]
    a_row_nnz = a_crow_inds[1:] - a_row_start
    a_col_inds = a.col_indices()
    a_vals = a.values()

    b_crow_inds = b.crow_indices()
    b_col_inds = b.col_indices()
    b_vals = b.values()
    b_row_start = b_crow_inds[:-1]
    b_row_nnz = b_crow_inds[1:] - b_row_start

    b_row_nnz_sorted, sorted_indices = torch.sort(b_row_nnz)

    b_row_start_sorted = b_row_start[sorted_indices]


    topk_inds = torch.empty(m, self.tpb, device=a.device, dtype=torch.long).fill_(-99)
    topk_vals = torch.empty(m, self.tpb, device=a.device, dtype=torch.float).fill_(float("-inf"))
    
    blocks_per_grid = (math.ceil(m / self.tile_m), )
    threads_per_block = (self.tpb, )
    self.fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      args=[
        a_row_start.data_ptr(),
        a_row_nnz.data_ptr(),
        a_col_inds.data_ptr(),
        a_vals.data_ptr(),

        b_row_start_sorted.data_ptr(),
        b_row_nnz_sorted.data_ptr(),
        b_col_inds.data_ptr(),
        b_vals.data_ptr(),
        
        topk_inds.data_ptr(),
        topk_vals.data_ptr(),
        m, n, k
      ],
      stream=self.stream,
      shared_mem=self.smem_size,
    )
    topk_vals = topk_vals[:, :n_candidates]
    topk_inds = topk_inds[:, :n_candidates]
    topk_inds = sorted_indices[topk_inds]

    return topk_vals, topk_inds 

  def call(
      self, 
      problem_size : ProblemSize,
      a_row_start : Tensor,
      a_row_nnz : Tensor,
      a_col_inds : Tensor,
      a_vals : Tensor,

      b_row_start : Tensor,  #[n]
      b_row_nnz : Tensor, #[n]
      b_col_inds : Tensor, #[n]
      b_vals : Tensor, #[n]

      k : int = 256,
    ):
    m_ = a_row_start.shape[0]

    sub_topk_inds = torch.empty(m_, self.tpb, device=a_row_start.device, dtype=torch.long).fill_(-99)
    sub_topk_vals = torch.empty(m_, self.tpb, device=a_row_start.device, dtype=torch.float).fill_(float("-inf"))
    
    blocks_per_grid = (math.ceil(m_ / self.tile_m), )
    threads_per_block = (self.tpb, )
    self.fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      args=[
        a_row_start.data_ptr(),
        a_row_nnz.data_ptr(),
        a_col_inds.data_ptr(),
        a_vals.data_ptr(),

        b_row_start.data_ptr(),
        b_row_nnz.data_ptr(),
        b_col_inds.data_ptr(),
        b_vals.data_ptr(),
        
        sub_topk_inds.data_ptr(),
        sub_topk_vals.data_ptr(),
        m_, problem_size.n, problem_size.k
      ],
      stream=self.stream,
      shared_mem=self.smem_size,
    )
    sub_topk_vals = sub_topk_vals[:, :k]
    sub_topk_inds = sub_topk_inds[:, :k]

    return sub_topk_vals, sub_topk_inds 


class TopkSPSPCSIM:
  def __init__(self, sim_type):
    # self.nnzpr_ranges = [0, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048]
    # self.tile_ms =      [4, 4, 4, 4, 4,  4,  4,  4,   3,   1,   1]
    # self.nnzpr_ranges = [0, 4, 16, 64, 256, 512, 1024, 2048]
    # self.tile_ms =      [   4, 4,  4,  4,   3,   1,    1]
    self.nnzpr_ranges = [0, 32, 256, 512, 1024, 2048]
    self.tile_ms =      [   4,  4,   3,   1,    1]
    self.tpbs = [128, 256, 512, 1024]
    self.timer = ProfilingTimer(False, name="TopkSPSPCSIM")
    for i in range(len(self.nnzpr_ranges) - 1):

      minnnzpr = self.nnzpr_ranges[i]
      maxnnzpr = self.nnzpr_ranges[i + 1]
      tile_m = self.tile_ms[i]
      for tpb in self.tpbs:
        fn = TopkSPSPCSIMCuda(
          tpb=tpb, 
          minnnzpr=minnnzpr, 
          maxnnzpr=maxnnzpr, 
          tile_m=tile_m, 
          threads_per_group=4, 
          sim_type=sim_type
        )
        setattr(self, f"fn_{tpb}_{minnnzpr}_{maxnnzpr}", fn)

  def __call__(self, a : Tensor, b : Tensor, k=128, n_candidates=None):
    assert isinstance(a, Tensor)
    assert isinstance(b, Tensor)
    assert a.ndim == b.ndim == 2
    assert a.shape[1] == b.shape[1]
    assert a.layout in { torch.sparse_csr, torch.sparse_coo}
    assert b.layout in { torch.sparse_csr, torch.sparse_coo}
    assert a.device == b.device
    assert a.dtype == b.dtype == torch.float
    if n_candidates is not None:
      k = n_candidates
    start, stop = self.timer.create_context("__call__")

    start("to sparse csr")
    assert 0 < k <= 1024
    if a.layout == torch.sparse_coo:
      a = a.to_sparse_csr()
    
    if b.layout == torch.sparse_coo:
      b = b.to_sparse_csr()
    stop("to sparse csr")
    # problem_size = (a.shape[0], *b.shape)
    start("init topk inds and values")
    problem_size = ProblemSize(m=a.shape[0], n=b.shape[0], k=b.shape[1])
    topk_inds = torch.empty(problem_size.m, k, device=a.device, dtype=torch.long).fill_(-99)
    topk_vals = torch.empty(problem_size.m, k, device=a.device, dtype=torch.float).fill_(float("-inf"))
    stop("init topk inds and values")
    
    start("get a crow col val")
    a_crow_inds = a.crow_indices()
    a_row_start = a_crow_inds[:-1]
    a_row_nnz = a_crow_inds[1:] - a_row_start
    a_col_inds = a.col_indices()
    a_vals = a.values()
    stop("get a crow col val")

    start("get b crow col val")
    b_crow_inds = b.crow_indices()
    b_col_inds = b.col_indices()
    b_vals = b.values()
    b_row_start = b_crow_inds[:-1]
    b_row_nnz = b_crow_inds[1:] - b_row_start
    sorted_b_row_nnz, sorted_indices = torch.sort(b_row_nnz)
    sorted_b_row_start = b_row_start[sorted_indices]
    stop("get b crow col val")
    
    if k <= 128:
      tpb = 128
    elif k <= 256:
      tpb = 256
    elif k <= 512:
      tpb = 512
    elif k <= 1024:
      tpb = 1024
    max_a_row_nnz = a_row_nnz.max().item()
    for i in range(len(self.nnzpr_ranges)-1):
      start("misc and mask")
      minnnzpr = self.nnzpr_ranges[i]
      maxnnzpr = self.nnzpr_ranges[i+1]
      if max_a_row_nnz <= minnnzpr:
        break
      fn = getattr(self, f"fn_{tpb}_{minnnzpr}_{maxnnzpr}")
      if i == len(self.nnzpr_ranges)-2:
        mask = (minnnzpr < a_row_nnz)
      else:
        mask = (minnnzpr <= a_row_nnz) & (a_row_nnz < maxnnzpr)
      selected_a_row_nnz = a_row_nnz[mask]
      stop("misc and mask")
      if len(selected_a_row_nnz) == 0:
        continue
      start("kernel")
      selected_a_row_start = a_row_start[mask]
      sub_topk_vals, sub_topk_inds = fn.call(
        problem_size,
        selected_a_row_start,
        selected_a_row_nnz,
        a_col_inds,
        a_vals,
        sorted_b_row_start,
        sorted_b_row_nnz,
        b_col_inds,
        b_vals,
        k
      )
      stop("kernel")
      start("sub_topk_vals to topk_vals")
      topk_vals[mask] = sub_topk_vals
      topk_inds[mask] = sub_topk_inds
      stop("sub_topk_vals to topk_vals")
    start("final map")
    topk_inds = sorted_indices[topk_inds]
    stop("final map")
    return topk_vals, topk_inds

  def __del__(self):
    self.timer.summarize()


from setuptools import setup, find_packages
from sparse_dok.__version__ import __version__
try:
  import cupy
except ImportError:
  print("Install cupy first.")
  exit()

setup(
  name = 'sparse_dok',
  packages = find_packages(),
  version = __version__,
  license='MIT',
  description = 'sparse dok tensor implementation',
  author = 'demoriarty', 
  author_email = 'sahbanjan@gmail.com',
  url = 'https://github.com/DeMoriarty/SparseDOK',
  download_url = f'https://github.com/DeMoriarty/SparseDOK/archive/v_{__version__.replace(".", "")}.tar.gz',
  keywords = ['pytorch', "sparse"],
  install_requires=[ 
    'numpy',
    'torch>=1.10.0',
    'sympy'
  ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
  ],
  include_package_data = True,
)


import torch
import numpy as np
import math
import sympy
from pathlib import Path
from torch import Tensor

from .cuda.closed_hashmap_impl import ClosedHashmapImplCuda
from ..util import str2dtype, next_power_of_2, unique_first_occurrence, expand_tensor, batch_allclose

class CudaClosedHashmap:
  def __init__(
      self, 
      n_buckets,
      device  = "cuda:0",
      prime1 = None,
      prime2 = None,
      alpha1 = None,
      alpha2 = None,
      beta1 = None,
      beta2 = None,
      key_size = None,
      key_perm = None,
      rehash_factor= 2.0,
      rehash_threshold = 0.75,
    ):
    self.n_buckets = n_buckets
    self.device = torch.device(device)
    self._uuid = torch.zeros(n_buckets, device=device, dtype=torch.long) - 1

    self._key_size = key_size
    self._keys = None
    self._values = None

    self._prime1 = prime1
    self._prime2 = prime2

    self._alpha1 = alpha1
    self._alpha2 = alpha2
    self._beta1 = beta1
    self._beta2 = beta2
    self._n_elements = torch.zeros(size=(), device=self.device, dtype=torch.long)
    self._rehash_factor = rehash_factor
    self._rehash_threshold = rehash_threshold

    self._unique_subkeys = None
    self._key_perm = None
    if key_perm is not None:
      if not isinstance(key_perm, Tensor):
        key_perm = torch.tensor(key_perm, dtype=torch.long, device=self.device)
      self._key_perm = key_perm.to(device=self.device, dtype=torch.long)
      # self._unique_subkeys = CudaClosedHashmap(32, device=self.device)

    self._hashmap_impl_cuda = None
    self.debug = False

  @property
  def hashmap_impl_cuda(self):
    if self._hashmap_impl_cuda is None:
      assert self._keys is not None
      self._hashmap_impl_cuda = ClosedHashmapImplCuda(
        tpb=256,
        kpt=1,
        key_size=np.product(*self.key_shape),
        value_size=np.product(*self.value_shape),
        key_type=self._keys.dtype,
        value_type=self._values.dtype,
      )
    return self._hashmap_impl_cuda

  @property
  def prime1(self):
    if self._prime1 is None:
      assert self._keys is not None or self._key_size is not None
      if self._keys is not None:
        key_size = self._keys[0].flatten().shape[0]
      elif self._key_size is not None:
        key_size = self._key_size
        
      self._prime1 = [np.random.randint(self.n_buckets, self.n_buckets*2) for i in range(key_size)]
      self._prime1 = torch.tensor([sympy.nextprime(i) for i in self._prime1], device=self.device)
    return self._prime1

  @property
  def prime2(self):
    if self._prime2 is None:
      assert self._keys is not None or self._key_size is not None
      if self._keys is not None:
        key_size = self._keys[0].flatten().shape[0]
      elif self._key_size is not None:
        key_size = self._key_size
      self._prime2 = [np.random.randint(0, 2**60, dtype="int64") for i in range(key_size)]
      self._prime2 = torch.tensor([sympy.nextprime(i) for i in self._prime2], device=self.device)
    return self._prime2

  @property
  def alpha1(self):
    if self._alpha1 is None:
      self._alpha1 = torch.tensor([np.random.randint(i.cpu()) - 1 for i in self.prime1], device=self.device)#torch.randint(0, size=self.prime1.shape)
    return self._alpha1
  
  @property
  def alpha2(self):
    if self._alpha2 is None:
      self._alpha2 = torch.tensor([np.random.randint(i.cpu()) - 1 for i in self.prime1], device=self.device)#torch.randint(0, size=self.prime1.shape)
    return self._alpha2
  
  @property
  def beta1(self):
    if self._beta1 is None:
      self._beta1 = torch.tensor([np.random.randint(i.cpu()) - 1 for i in self.prime1], device=self.device)#torch.randint(0, size=self.prime1.shape)
    return self._beta1
  
  @property
  def beta2(self):
    if self._beta2 is None:
      self._beta2 = torch.tensor([np.random.randint(i.cpu()) - 1 for i in self.prime1], device=self.device)#torch.randint(0, size=self.prime1.shape)
    return self._beta2

  @property
  def key_perm(self):
    if self._key_perm is None:
      assert self._keys is not None
      self._key_perm = torch.arange(np.prod(self.key_shape), device=self.device)
    else:
      assert self._key_perm.shape[0] == np.prod(self.key_shape)
    return self._key_perm

  @key_perm.setter
  def key_perm(self, new):
    if not isinstance(new, Tensor):
      new = torch.tensor(new, device=self.device, dtype=torch.long)
      assert new.shape == self.key_perm.shape
      self._key_perm = new

  def keys(self):
    if self._keys is None:
      return None
    else:
      mask = self._uuid != -1
      keys = self._keys.reshape(self._keys.shape[0], -1)[mask][:, self.key_perm]
      return keys

  def values(self):
    if self._values is None:
      return None
    else:
      mask = self._uuid != -1
      return self._values[mask]

  def set_values(self, selectors, new_values):
    mask = self._uuid != -1
    masked_values = self._values[mask]
    masked_values[selectors] = new_values
    self._values[mask] = masked_values

  def permute_keys(self, ordering):
    assert self._keys is not None, "hashmap is empty"
    assert len(self.key_shape) == 1, "keys are not one dimensional"
    assert len(ordering) == self.key_shape[0]
    assert len(set(ordering)) == len(ordering), "duplicates are not allowed"
    self._keys = self._keys[:, ordering]
    self._alpha1 = self._alpha1[ordering] 
    self._alpha2 = self._alpha2[ordering] 
    self._beta1 = self._beta1[ordering] 
    self._beta2 = self._beta2[ordering] 
    self._prime1 = self._prime1[ordering] 
    self._prime2 = self._prime2[ordering]
    self._key_size = np.prod(self._keys.shape[1:])

  def clone(self, deepclone=False):
    if deepclone:
      new = CudaClosedHashmap(
        n_buckets=self.n_buckets, 
        device=self.device,
        prime1=self.prime1.clone(),  
        prime2=self.prime2.clone(), 
        alpha1=self.alpha1.clone(), 
        alpha2=self.alpha2.clone(), 
        beta1=self.beta1.clone(), 
        beta2=self.beta2.clone(),
        key_perm=self._key_perm.clone(),
        rehash_factor=self._rehash_factor,
        rehash_threshold=self._rehash_threshold,
      )
      new._keys = self._keys.clone()
      new._values = self._values.clone()
      new._uuid = self._uuid.clone()
      new._n_elements = self._n_elements.clone()
      new._hashmap_impl_cuda = self._hashmap_impl_cuda
      return new
    else:
      new = CudaClosedHashmap(
        n_buckets=self.n_buckets, 
        device=self.device,
        prime1=self.prime1,  
        prime2=self.prime2, 
        alpha1=self.alpha1, 
        alpha2=self.alpha2, 
        beta1=self.beta1, 
        beta2=self.beta2,
        key_perm=self._key_perm,
        rehash_factor=self._rehash_factor,
        rehash_threshold=self._rehash_threshold,
      )
      new._keys = self._keys
      new._values = self._values
      new._uuid = self._uuid
      new._n_elements = self._n_elements
      new._hashmap_impl_cuda = self._hashmap_impl_cuda
      return new

  def uuid(self):
    mask = self._uuid != -1
    return self._uuid[mask]

  @property
  def n_elements(self):
    return self._n_elements.item()

  @property
  def load_factor(self):
    return self.n_elements / self.n_buckets

  @property
  def key_shape(self):
    assert self._keys is not None
    return self._keys.shape[1:]
  
  @property
  def value_shape(self):
    assert self._values is not None
    return self._values.shape[1:]
    
  def test_uniformity(self, keys):
    hash_codes = self.get_hash(keys)
    logits = (torch.bincount(hash_codes, minlength=self.n_buckets) + 1).log()
    def entropy(logits):
      logp = torch.log_softmax(logits, dim=-1)
      p = torch.softmax(logits, dim=-1)
      return -(logp * p).sum(dim=-1)
    return entropy(logits).exp() / self.n_buckets

  def __len__(self):
    return self.n_elements

  def get_hash(self, keys): #[n_keys, d_key]):
    # torch.quantize_per_tensor(torch.randn(1000,8), 0.1, 0, torch.qint32).int_repr().sum(-1) % 10
    n_keys = keys.shape[0]
    keys = keys.reshape(n_keys, -1).clone()
    if keys.dtype == torch.float:
      keys = keys.view(dtype=torch.int32).long()
      ### swap -0.0 with 0.0
      keys[keys == -2**31] = 0
    elif keys.dtype == torch.half:
      keys = keys.view(dtype=torch.int16).long()
      keys[keys == -2**15] = 0
    elif keys.dtype == torch.double:
      keys = keys.view(dtype=torch.int64)
      keys[keys == -2**63] = 0
    else:
      keys = keys.to(torch.long)
    
    # modified from https://www.youtube.com/watch?v=Kf2V77ut-B0
    if self._key_perm is None:
      h = ( (self.alpha1[None] * keys + self.beta1[None]) % self.prime1[None] ).prod(dim=-1).abs() % self.n_buckets
    else:
      h = ( (self.alpha1[None] * keys + self.beta1[None]) % self.prime1[None] )[:, self._key_perm].prod(dim=-1).abs() % self.n_buckets
    return h

  def get_uuid(self, keys): #[n_keys, d_key]):
    n_keys = keys.shape[0]
    keys = keys.reshape(n_keys, -1).clone()
    if keys.dtype == torch.float:
      keys = keys.view(dtype=torch.int32).long()
      ### swap -0.0 with 0.0
      keys[keys == -2**31] = 0
    elif keys.dtype == torch.half:
      keys = keys.view(dtype=torch.int16).long()
      keys[keys == -2**15] = 0
    elif keys.dtype == torch.double:
      keys = keys.view(dtype=torch.int64)
      keys[keys == -2**63] = 0
    else:
      keys = keys.to(torch.long)
    
    h = ( (self.alpha2[None] * keys + self.beta2[None]) % self.prime2[None] ).prod(dim=-1).abs()
    return h

  def set(self, new_keys, new_values):
    assert new_keys.shape[0] == new_values.shape[0]

    if self._keys is None:
      self._keys = torch.zeros(
        self.n_buckets,
        *new_keys.shape[1:],
        device=self.device,
        dtype=new_keys.dtype
      )
      self._values = torch.zeros(
        self.n_buckets,
        *new_values.shape[1:],
        device=self.device,
        dtype=new_values.dtype,
      )
    else:
      assert new_keys.dtype == self._keys.dtype
      assert new_keys.device == self._keys.device == self.device
      assert new_keys.shape[1:] == self._keys.shape[1:]
      assert new_values.dtype == self._values.dtype
      assert new_values.device == self._values.device == self.device
      assert new_values.shape[1:] == self._values.shape[1:]
    new_keys = new_keys.contiguous()
    new_values = new_values.contiguous()

    if new_keys.shape[0] == 0:
      return

    _, is_found = self.hashmap_impl_cuda.get(
      prime1=self.prime1,
      prime2=self.prime2,
      alpha1=self.alpha1,
      alpha2=self.alpha2,
      beta1=self.beta1,
      beta2=self.beta2,
      key_perm=self.key_perm,
      keys=new_keys,
      all_keys=self._keys,
      all_values=self._values,
      all_uuids=self._uuid
    )
    n_new_elements = new_keys[~is_found].unique(dim=0).shape[0]

    if self.n_elements + n_new_elements > int(self._rehash_threshold * self.n_buckets):
      self.rehash( int((self.n_elements + n_new_elements) * self._rehash_factor) )

    is_stored = self.hashmap_impl_cuda.set(
      prime1=self.prime1,
      prime2=self.prime2,
      alpha1=self.alpha1,
      alpha2=self.alpha2,
      beta1=self.beta1,
      beta2=self.beta2,
      key_perm=self.key_perm,
      keys=new_keys,
      values=new_values,
      all_keys=self._keys,
      all_values=self._values,
      all_uuids=self._uuid
    )


    self._n_elements += n_new_elements

  def get(
      self,
      keys,
      fallback_value = None,
    ):
    if self._keys is None or self._values is None:
      raise RuntimeError("store something first")
      return None, None

    if keys.shape[0] == 0:
      return torch.empty(0, *self.value_shape, device=self.device, dtype=self._values.dtype), torch.empty(0, device=self.device, dtype=torch.bool)
    assert keys.dtype == self._keys.dtype
    assert keys.device == self._keys.device
    assert keys.shape[1:] == self.key_shape
    if fallback_value is not None:
      if not isinstance(fallback_value, torch.Tensor):
        fallback_value = torch.tensor(fallback_value, device=self.device, dtype=self._values.dtype)
      fallback_value = torch.broadcast_to(fallback_value, self.value_shape)
    keys = keys.contiguous()
    values, is_found = self.hashmap_impl_cuda.get(
      prime1=self.prime1,
      prime2=self.prime2,
      alpha1=self.alpha1,
      alpha2=self.alpha2,
      beta1=self.beta1,
      beta2=self.beta2,
      key_perm=self.key_perm,
      keys=keys,
      all_keys=self._keys,
      all_values=self._values,
      all_uuids=self._uuid,
      fallback_value=fallback_value,
    )
    # print(values)
    # print(is_found.sum(), is_found.numel())
    return values, is_found

  def rehash(self, n_buckets):
    # print(f"rehashing {n_buckets}")
    keys = self.keys().contiguous()
    values = self.values().contiguous()
    
    self.__init__(
      n_buckets=n_buckets,
      device=self.device,
      key_perm=self._key_perm,
      rehash_factor=self._rehash_factor,
      rehash_threshold=self._rehash_threshold,
    )
    self.set(keys, values)

  def __getitem__(self, keys):
    return self.get(keys)

  def __setitem__(self, keys, values):
    return self.set(keys, values)


import cupy as cp
import torch
import math
from torch import Tensor

from ..cuda_callable import CudaCallable
from ...util import get_absolute_path, dtype2ctype, str2dtype

class ClosedHashmapImplCuda(CudaCallable):
  def __init__(
      self, 
      tpb=256, 
      kpt=1, 
      key_size=1, 
      key_type="long", 
      value_size=1, 
      value_type="float",
    ):
    super().__init__()
    assert 32 <= tpb <= 1024
    assert kpt >= 1
    assert key_size >= 1 
    assert value_size >= 1

    self.tpb = tpb
    self.kpt = kpt
    self.torch_key_type = str2dtype(key_type)
    self.torch_value_type = str2dtype(value_type)
    self.key_type = dtype2ctype(key_type)
    self.value_type = dtype2ctype(value_type)
    self.key_size = key_size
    self.value_size = value_size

    cu_files = [
      "head.cu",
      "smem_tensor.cu",
      "closed_hashmap_impl.cu",
    ]

    kernel = []
    for file in cu_files:
      with open(get_absolute_path("components", "cuda", file), "r") as f:
        kernel.append(f.read())

    self.kernel = "\n".join(kernel)

    with open(get_absolute_path("components", "cuda", "preview.cu"), "w") as f:
      f.write(self.kernel)

    self.kernel = (self.kernel
      .replace("_TPB_", str(self.tpb))
      .replace("_KPT_", str(self.kpt))
      .replace("_KEYSIZE_", str(self.key_size))
      .replace("_VALUESIZE_", str(self.value_size))
      .replace("_KEYTYPE_", str(self.key_type))
      .replace("_VALUETYPE_", str(self.value_type))
    )

    self._get_fn = cp.RawKernel(
      self.kernel,
      "closed_hashmap_get",
      backend="nvcc",
      options=(
        '-std=c++17',
        "--device-c",
      )
    )

    self._set_fn = cp.RawKernel(
      self.kernel,
      "closed_hashmap_set",
      backend="nvcc",
      options=(
        '-std=c++17',
        "--device-c",
      )
    )
    # self.fn.max_dynamic_shared_size_bytes = self.smem_size
  
  def get(
      self,
      prime1,
      prime2,
      alpha1,
      alpha2,
      beta1,
      beta2,
      key_perm ,
      keys, #[n_keys, key_size]
      all_keys, #[n_all_keys, key_size]
      all_values, #[n_all_keys, value_size]
      all_uuids, #[n_all_keys]
      values = None, #[n_keys, value_size]
      fallback_value = None, #[value_size]
    ):
    assert keys.ndim == all_keys.ndim == all_values.ndim == 2
    assert all_keys.shape[1] == keys.shape[1] == self.key_size
    assert all_values.shape[1] == self.value_size
    assert keys.dtype == all_keys.dtype == self.torch_key_type
    assert all_values.dtype == self.torch_value_type
    assert all_keys.shape[0] == all_values.shape[0] == all_uuids.shape[0] != 0
    assert keys.device == all_keys.device == all_values.device == all_uuids.device == self.device
    assert alpha1.shape == beta1.shape == prime1.shape == alpha2.shape == beta2.shape == prime2.shape == (self.key_size, )
    assert alpha1.dtype == beta1.dtype == prime1.dtype == alpha2.dtype == beta2.dtype == prime2.dtype == torch.long
    assert alpha1.device == beta1.device == prime1.device == alpha2.device == beta2.device == prime2.device == self.device
    assert keys.is_contiguous() and all_keys.is_contiguous() and all_values.is_contiguous() and all_uuids.is_contiguous()
    assert key_perm.shape == alpha1.shape
    assert key_perm.device == alpha1.device
    assert key_perm.dtype == torch.long

    
    n_keys = keys.shape[0]
    n_buckets = all_keys.shape[0]
    if fallback_value is not None:
      assert fallback_value.dtype == self.torch_value_type
      assert fallback_value.ndim == 1
      assert fallback_value.device == self.device
      assert fallback_value.shape[0] == self.value_size
    else:
      fallback_value = torch.zeros(
        self.value_size,
        device=keys.device,
        dtype=self.torch_value_type  
      )

    if values is not None:
      assert values.dtype == self.torch_value_type
      assert values.device == self.device
      assert values.ndim == 2
      assert values.shape[0] == keys.shape[0]
      assert values.shape[1] == self.value_size
      assert values.is_contiguous()
    else:
      values = torch.empty(
        n_keys,
        self.value_size,
        device=keys.device,
        dtype=self.torch_value_type
      )
      values[:] = fallback_value[None]

    is_found = torch.zeros(n_keys, device=self.device, dtype=torch.bool)

    blocks_per_grid = ( math.ceil(n_keys / (self.tpb * self.kpt)), )
    threads_per_block = (self.tpb, )
    self._get_fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      args=[
        prime1.data_ptr(),
        prime2.data_ptr(),
        alpha1.data_ptr(), 
        alpha2.data_ptr(),
        beta1.data_ptr(),
        beta2.data_ptr(),
        key_perm.data_ptr(),
        keys.data_ptr(), #[n_keys, key_size]
        values.data_ptr(), #[n_keys, value_size]
        all_keys.data_ptr(), #[n_all_keys, key_size]
        all_values.data_ptr(), #[n_all_keys, value_size]
        all_uuids.data_ptr(), #[n_all_keys]
        fallback_value.data_ptr(), #[value_size]
        is_found.data_ptr(),
        n_keys, n_buckets,
      ]
    )

    return values, is_found

  def set(
      self,
      prime1,
      prime2,
      alpha1,
      alpha2,
      beta1,
      beta2,
      key_perm,
      keys, #[n_keys, key_size]
      values, #[n_keys, value_size]
      all_keys, #[n_all_keys, key_size]
      all_values, #[n_all_keys, value_size]
      all_uuids, #[n_all_keys]
    ):
    assert keys.ndim == values.ndim == all_keys.ndim == all_values.ndim == 2
    assert values.shape[0] == keys.shape[0]
    assert all_keys.shape[1] == keys.shape[1] == self.key_size
    assert all_values.shape[1] == values.shape[1] == self.value_size
    assert keys.dtype == all_keys.dtype == self.torch_key_type
    assert values.dtype == all_values.dtype == self.torch_value_type
    assert all_keys.shape[0] == all_values.shape[0] == all_uuids.shape[0] != 0
    assert keys.device == values.device == all_keys.device == all_values.device == all_uuids.device == self.device
    assert alpha1.shape == beta1.shape == prime1.shape == alpha2.shape == beta2.shape == prime2.shape == (self.key_size, )
    assert alpha1.dtype == beta1.dtype == prime1.dtype == alpha2.dtype == beta2.dtype == prime2.dtype == torch.long
    assert alpha1.device == beta1.device == prime1.device == alpha2.device == beta2.device == prime2.device == self.device
    assert key_perm.shape == alpha1.shape
    assert key_perm.device == alpha1.device
    assert key_perm.dtype == torch.long
    assert keys.is_contiguous() and values.is_contiguous() and all_keys.is_contiguous() and all_values.is_contiguous() and all_uuids.is_contiguous()

    n_keys = keys.shape[0]
    n_buckets = all_keys.shape[0]
    is_stored = torch.zeros(n_keys, device=self.device, dtype=torch.bool)

    blocks_per_grid = ( math.ceil(n_keys / (self.tpb * self.kpt)), )
    threads_per_block = (self.tpb, )
    self._set_fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      args=[
        prime1.data_ptr(),
        prime2.data_ptr(),
        alpha1.data_ptr(), 
        alpha2.data_ptr(),
        beta1.data_ptr(),
        beta2.data_ptr(),
        key_perm.data_ptr(),
        keys.data_ptr(), #[n_keys, key_size]
        values.data_ptr(), #[n_keys, value_size]
        all_keys.data_ptr(), #[n_all_keys, key_size]
        all_values.data_ptr(), #[n_all_keys, value_size]
        all_uuids.data_ptr(), #[n_all_keys]
        is_stored.data_ptr(),
        n_keys, n_buckets,
      ]
    )
    return is_stored